@@ -87,7 +87,10 @@ class MicroManagerV1Dataset(
         return shape, dtype
 
     def _load_channel_info(self):
-        return self.metadata["summary"]["ChNames"]
+        channels = self.metadata["summary"]["ChNames"]
+        if len(channels) == 1 and channels[0] == "Default":
+            channels = ["*"]
+        return channels
 
     def _load_metadata(self, filename="metadata.txt"):
         # find all `metadata.txt` and try to open until success
@@ -102,7 +105,9 @@ class MicroManagerV1Dataset(
                     metadata = {}
                     metadata["summary"] = raw_metadata["Summary"]
                     for key, value in raw_metadata.items():
-                        if key.startswith("Metadata") and key.endswith(".tif"):
+                        if (
+                            key.startswith("Metadata") and key.endswith(".tif")
+                        ) or key.startswith("FrameKey-"):
                             metadata["frame"] = value
                             break
                     else:
@@ -155,103 +160,44 @@ class MicroManagerV1Dataset(
             metadata["frame"] = next(iter(ijson.items(fd, frame_metadata_key)))
         return metadata
 
-    def _load_mapped_coordinates(self):  # TODO update to new format
+    def _parse_position_list(self):
         positions = self.metadata["summary"]["InitialPositionList"]
 
-        coords = defaultdict(list)
-        labels = dict()
-        for position in positions:
-            # coordinate
-            coord_dict = position["DeviceCoordinatesUm"]
-            try:
-                coord_x, coord_y = tuple(coord_dict["XY Stage"])
-                coord_x, coord_y = np.float32(coord_x), np.float32(coord_y)
+        frame = self.metadata["frame"]
+        xy_stage, z_stage = frame["Core-XYStage"], frame["Core-Focus"]
+        logger.debug(f'XY stage is "{xy_stage}", Z stage is "{z_stage}"')
 
-                coords["tile_x"].append(coord_x)
-                coords["tile_y"].append(coord_y)
-            except KeyError:
-                pass
-            try:
-                coord_z = coord_dict["Z Stage"][0]
-                coord_z = np.float32(coord_z)
-
-                coords["tile_z"].append(coord_z)
-            except KeyError:
-                pass
-
-            # label
-            # NOTE MicroManager only tiles in 2D, no need to include Z for indexing
-            labels[(coord_x, coord_y)] = position["Label"]
-
-        # internal bookkeeping
-        self._tile_prefix = labels
-
-        return pd.DataFrame({k: np.array(v) for k, v in coords.items()})
-
-    def _load_voxel_size(self):
-        dx, r = (
-            self.metadata["summary"]["PixelSize_um"],
-            self.metadata["summary"]["PixelAspect"],
-        )
-        if dx == 0:
-            dx = prompt_pixel_size()
-        size = (r * dx, dx)
-
-        if self.metadata["summary"]["Slices"] > 1:
-            size = (abs(self.metadata["summary"]["z-step_um"]),) + size
-
-        return size
-
-    # def _missing_data(self):
-    #    shape, dtype = self._load_array_info()
-    #    return delayed(np.zeros)(shape, dtype)
-
-    def _retrieve_file_list(self, coord_dict):
-        prefix = self._tile_prefix[itemgetter("tile_x", "tile_y")(coord_dict)]
-        return glob.glob(
-            os.path.join(self.root_dir, prefix, f"*_{coord_dict['channel']}_*.tif")
-        )
-
-
-class MicroManagerV2Dataset(MicroManagerV1Dataset):
-    def _can_read(self):
-        version = self.metadata["summary"]["MicroManagerVersion"]
-        return version.startswith("2.")
-
-    def _load_channel_info(self):
-        channels = self.metadata["summary"]["ChNames"]
-        if len(channels) == 1 and channels[0] == "Default":
-            channels = ["*"]
-
-        # internal bookkeeping
-        self._channel_order = channels
-
-        return channels
-
-    def _load_mapped_coordinates(self):
         coords, index = [], []
         labels = []
-        for position in self.metadata["summary"]["StagePositions"]:
+        for position in positions:
             # NOTE MicroManager only tiles in 2D
-            index.append({"x": position["GridCol"], "y": position["GridRow"]})
+            index.append(
+                {"x": position["GridColumnIndex"], "y": position["GridRowIndex"]}
+            )
 
-            xy_stage, z_stage = position["DefaultXYStage"], position["DefaultZStage"]
+            # coordinate
+            coord_dict = position["DeviceCoordinatesUm"]
             coord = {}
-            for device in position["DevicePositions"]:
-                device_name = device["Device"]
-                if device_name == xy_stage:
-                    coord_x, coord_y = tuple(device["Position_um"])
-                    coord["x"] = float(coord_x)
-                    coord["y"] = float(coord_y)
-                elif device_name == z_stage:
-                    coord_z = device["Position_um"][0]
-                    coords["z"] = float(coord_z)
-                else:
-                    logger.warning(f"unknown device {device_name}")
+            try:
+                coord_x, coord_y = tuple(coord_dict[xy_stage])
+                coord["x"] = float(coord_x)
+                coord["y"] = float(coord_y)
+            except KeyError:
+                pass
+            try:
+                coord_z = coord_dict[z_stage][0]
+                coord["z"] = float(coord_z)
+            except KeyError:
+                pass
             coords.append(coord)
 
             labels.append(position["Label"])
 
+        return coords, index, labels
+
+    def _load_mapped_coordinates(self):
+        coords, index, labels = self._parse_position_list()
+
         coords, index = pd.DataFrame(coords), pd.DataFrame(index)
         labels = pd.DataFrame({"label": labels})
 
@@ -273,6 +219,81 @@ class MicroManagerV2Dataset(MicroManagerV1Dataset):
 
         return df
 
+    def _load_voxel_size(self):
+        dx, r = (
+            self.metadata["summary"]["PixelSize_um"],
+            self.metadata["summary"]["PixelAspect"],
+        )
+        if dx == 0:
+            dx = prompt_pixel_size()
+        size = (r * dx, dx)
+
+        if self.metadata["summary"]["Slices"] > 1:
+            size = (abs(self.metadata["summary"]["z-step_um"]),) + size
+
+        return size
+
+    # def _missing_data(self):
+    #    shape, dtype = self._load_array_info()
+    #    return delayed(np.zeros)(shape, dtype)
+
+    def _retrieve_file_list(self, coord_dict):
+        # find folder that contains the stack
+        tile_index = ["tile_x", "tile_y"]
+        label = self.tile_coords.xs(
+            itemgetter(*tile_index)(coord_dict), axis="index", level=tile_index,
+        )["label"].iloc[0]
+        tile_folder = os.path.join(self.root_dir, label)
+        file_list = [f for f in self.files if f.startswith(tile_folder)]
+
+        # find images
+        ch = coord_dict["channel"]
+        if ch != "*":
+            # we need to match channels explicitly
+            file_list = [f for f in file_list if ch in f]
+
+        return file_list
+
+
+class MicroManagerV2Dataset(MicroManagerV1Dataset):
+    def _can_read(self):
+        version = self.metadata["summary"]["MicroManagerVersion"]
+        return version.startswith("2.")
+
+    def _load_channel_info(self):
+        channels = super()._load_channel_info()
+
+        # internal bookkeeping
+        self._channel_order = channels
+
+        return channels
+
+    def _parse_position_list(self):
+        coords, index = [], []
+        labels = []
+        for position in self.metadata["summary"]["StagePositions"]:
+            # NOTE MicroManager only tiles in 2D
+            index.append({"x": position["GridCol"], "y": position["GridRow"]})
+
+            xy_stage, z_stage = position["DefaultXYStage"], position["DefaultZStage"]
+            coord = {}
+            for device in position["DevicePositions"]:
+                device_name = device["Device"]
+                if device_name == xy_stage:
+                    coord_x, coord_y = tuple(device["Position_um"])
+                    coord["x"] = float(coord_x)
+                    coord["y"] = float(coord_y)
+                elif device_name == z_stage:
+                    coord_z = device["Position_um"][0]
+                    coords["z"] = float(coord_z)
+                else:
+                    logger.warning(f"unknown device {device_name}")
+            coords.append(coord)
+
+            labels.append(position["Label"])
+
+        return coords, index, labels
+
     def _load_voxel_size(self):
         frame_metadata = self.metadata["frame"]
         matrix = frame_metadata["PixelSizeAffine"]
@@ -288,10 +309,12 @@ class MicroManagerV2Dataset(MicroManagerV1Dataset):
         binning = frame_metadata["Binning"]
         binning = int(binning)
 
-        # calculate affine matrix
+        # affine matrix
         #   [ 1.0, 0.0, 0.0; 0.0, 1.0, 0.0 ]
-        # NOTE we don't care about directions at this stage
-        size = (abs(matrix[4] * binning), abs(matrix[0] * binning))
+        # NOTE
+        #   - we don't care about directions at this stage
+        #   - MM already factor in binning to the matrix
+        size = (abs(matrix[4]), abs(matrix[0]))
         logger.info(f"binning {binning}x, effective pixel size {size} um")
 
         if self.metadata["summary"]["Slices"] > 1:
