@@ -1,6 +1,7 @@
 import logging
 import os
 from collections import defaultdict
+from itertools import chain
 from typing import List, Optional
 
 import dask.array as da
@@ -11,7 +12,7 @@ import zarr
 from dask import delayed
 from dask.distributed import as_completed
 
-from utoolbox.utils.dask import get_client, wait_futures
+from utoolbox.util.dask import get_client
 
 from ..base import (
     DenseDataset,
@@ -54,12 +55,15 @@ class ZarrDataset(
         label (str, optional): label of the data array
         level (int, optional): resolution level
         path (str, optional): internal group path
+        strict (bool, optional): resolution level should exist instead of fallback
     """
 
     def __init__(
-        self, store: str, label: str = RAW_DATA_LABEL, level: int = 0, path: str = "/",
+        self, store: str, label: str = RAW_DATA_LABEL, level: int = 0, path: str = "/"
     ):
-        self._label, self._level = label, 0
+        if level < 0:
+            raise ValueError("resolution level should >= 0")
+        self._label, self._level = label, level
 
         super().__init__(store, path)
 
@@ -76,14 +80,41 @@ class ZarrDataset(
 
     @level.setter
     def level(self, level: int):
+        if level < 0:
+            raise ValueError("level should >= 0")
+        self._level = level
+
         # TODO how to reload dataset from filesystem
         raise NotImplementedError
 
     @property
     def read_func(self):
         def func(uri, shape, dtype):
+            group = self.handle[uri]
+
+            # multi-scale?
+            if ZarrDataset.is_multiscales(group):
+                name = os.path.basename(uri)
+                for group_attrs in group.attrs["multiscales"]:
+                    # find the corresponding attributes
+                    if group_attrs["name"] == name:
+                        # get path for requested level
+                        level = group_attrs["datasets"][self.level]["path"]
+                        break
+                else:
+                    raise RuntimeError(
+                        f'corrupted multiscale dataset, unable to find path for "{name}" (level: {self.level})'
+                    )
+            else:
+                level = ""
+
+            # build final path
+            # NOTE "zarr.Group + sub-path" does not function properly, use "str + full
+            # path" instead
+            path = "/".join([uri, level])
+
             # zarr array contains shape, dtype and decompression info
-            return da.from_zarr(self.root_dir, uri)
+            return da.from_zarr(self.root_dir, path)
 
         return func
 
@@ -171,7 +202,7 @@ class ZarrDataset(
                     tiles_iterator = TiledDatasetIterator(
                         v_selected, axes="zyx", return_format="both"
                     )
-                    for st, selected in tiles_iterator:
+                    for (st_index, st_coord), selected in tiles_iterator:
                         s_root = c_root.require_group(f"s{i_s}")
                         i_s += 1
 
@@ -185,32 +216,30 @@ class ZarrDataset(
                             s_root.attrs["view"] = sv
 
                         # tile attribute
-                        if st is None:
+                        if st_index is None:
                             for index in ("tile_index", "tile_coord"):
                                 try:
                                     del s_root.attrs[index]
                                 except KeyError:
                                     pass
                         else:
-                            index, coord = st
                             names = tiles_iterator.index
                             names = [name.split("_")[-1] for name in names]
                             # NOTE numpy dtypes are not serializable, use native
                             s_root.attrs["tile_index"] = {
-                                k: int(v) for k, v in zip(names, index)
+                                k: int(v) for k, v in zip(names, st_index)
                             }
                             s_root.attrs["tile_coord"] = {
-                                k: float(v) for k, v in zip(names, coord)
+                                k: float(v) for k, v in zip(names, st_coord)
                             }
 
-                        ## WIP
-
                         # 4-1) retrieve info
                         data = dataset[selected]
 
                         # 4-2) label
+                        write_back = True
                         if label in s_root:
-                            lbl_group = s_root[label]
+                            label_group = s_root[label]
 
                             # get the hash
                             hash_gen.update(data.compute())
@@ -218,65 +247,68 @@ class ZarrDataset(
                             hash_gen.reset()
 
                             try:
-                                dst_hash = lbl_group.attrs["checksum"]
+                                dst_hash = label_group.attrs["checksum"]
                                 if dst_hash != src_hash:
                                     raise ValueError
                             except KeyError:
                                 # hash does not exist, since it is only updated when
                                 # dump was completed
                                 logger.warning(
-                                    f'"{lbl_group.path}" contains partial dump, rewrite'
+                                    f'"{label_group.path}" contains partial dump, rewrite'
                                 )
                             except ValueError:
                                 # hash mismatch
                                 logger.warning(
                                     "existing hash does not match the source"
                                 )
+                                # reset
+                                del label_group.attrs["checksum"]
                             else:
                                 # array exists, and checksum matches
                                 if not overwrite:
-                                    logger.info(f'"{lbl_group.path}" already exists')
-                                    continue
+                                    logger.info(f'"{label_group.path}" already exists')
+                                    write_back = False
                         else:
                             # never seen this label before, create new one
                             # NOTE raw data will always be multiscale
-                            lbl_group = s_root.require_group(label)
-                            lbl_group.attrs["voxel_size"] = dataset.voxel_size
+                            label_group = s_root.require_group(label)
+                            label_group.attrs["voxel_size"] = dataset.voxel_size
 
                         # 5) level
                         # NOTE comply with multiscale arrays v0.1
                         # https://forum.image.sc/t/multiscale-arrays-v0-1/37930
 
                         # drop current multiscale attributes
-                        if "multiscales" in lbl_group.attrs:
+                        if "multiscales" in label_group.attrs:
                             # delete all existing levels
-                            multiscales = lbl_group.attrs["multiscales"]
-                            for multiscale in multiscales:
-                                for path in multiscale["datasets"]:
-                                    del lbl_group[path["path"]]
+                            multiscales = label_group.attrs["multiscales"]
+                            if multiscales:
+                                logger.warning("deleting existing multiscale datasets")
+                                for multiscale in multiscales:
+                                    for path in multiscale["datasets"]:
+                                        try:
+                                            del label_group[path["path"]]
+                                        except KeyError:
+                                            logger.warning(
+                                                f'"{path}" is already deleted'
+                                            )
 
-                        # regenerate
-                        multiscales = []
-                        if level > 0:
-                            logger.error(
-                                "on-the-fly generate level pyramid is not supported yet"
-                            )
-                        for lvl in levels:
-                            pass
-
-                        # multiscale = {"name": label, "datasets": [], "version": "0.1"}
-                        level = str(level)
-                        if level in lbl_group:
-                            data_dst = lbl_group[level]
-                            # generate before, but not this resolution level
-                            raise NotImplementedError
-                            # TODO ask user to regenerate other resolution level or not
+                        # generate 0-level, single-level only
+                        level_str = str("0")
+                        multiscales = [
+                            {
+                                "name": label,
+                                "datasets": [{"path": level_str}],
+                                "version": "0.1",
+                            }
+                        ]
+                        if level_str in label_group:
+                            data_dst = label_group[level_str]
                         else:
-                            # create new array
                             # NOTE compression benchmark reference http://alimanfoo.
                             # github.io/2016/09/21/genotype-compression-benchmark.html
-                            data_dst = lbl_group.empty_like(
-                                level,
+                            data_dst = label_group.empty_like(
+                                level_str,
                                 data,
                                 chunks=True,
                                 compression="blosc",
@@ -284,64 +316,11 @@ class ZarrDataset(
                                     cname="lz4", clevel=5, shuffle=zarr.blosc.SHUFFLE
                                 ),
                             )
+                        label_group.attrs["multiscales"] = multiscales
 
-                        ## WIP
-
-                        ## >>> TO REFACTOR
-
-                        # 4) level # FIXME swap order with label, label -> level
-                        l0_group = s_root.require_group("0")
-                        l0_group.attrs["voxel_size"] = dataset.voxel_size
-
-                        # 5) data
-                        logger.debug(f'writing "{l0_group.path}"')
-                        data = dataset[selected]
-
-                        if label in l0_group:
-                            # get the hash
-                            hash_gen.update(data.compute())
-                            src_hash = hash_gen.hexdigest()
-                            hash_gen.reset()
-
-                            try:
-                                dst_hash = l0_group.attrs["checksum"]
-                                if dst_hash != src_hash:
-                                    raise ValueError
-                            except KeyError:
-                                # hash does not exist, since it is only updated when
-                                # dump was complete
-                                logger.warning("last dump was incomplete")
-                            except ValueError:
-                                # hash mismatch
-                                logger.warning(
-                                    "existing hash does not match the source"
-                                )
-                            else:
-                                # array exists, and checksum matches
-                                if not overwrite:
-                                    logger.info(
-                                        f'"{l0_group.path}" already has "{label}"'
-                                    )
-                                    continue
-
-                            # select existing array
-                            data_dst = l0_group[label]
-                        else:
-                            # create new array
-
-                            # NOTE compression benchmark reference http://alimanfoo.
-                            # github.io/2016/09/21/genotype-compression-benchmark.html
-                            data_dst = l0_group.empty_like(
-                                label,
-                                data,
-                                chunks=True,
-                                compression="blosc",
-                                compression_opts=dict(
-                                    cname="lz4", clevel=5, shuffle=zarr.blosc.SHUFFLE
-                                ),
-                            )
-
-                        ## <<< TO REFACTOR
+                        # complete updates all the attributes, early stop here
+                        if not write_back:
+                            continue
 
                         # NOTE using default chunk shape after rechunk will cause
                         # problem, since chunk size is composite of chunks as tuples
@@ -362,7 +341,7 @@ class ZarrDataset(
 
                         checksum = calc_checksum(array)
                         future = client.compute(checksum)
-                        futures[future] = l0_group
+                        futures[future] = label_group
                         # TODO add callback here to update progressbar
 
         if not futures:
@@ -383,6 +362,46 @@ class ZarrDataset(
 
     ##
 
+    @staticmethod
+    def is_multiscales(group, strict=False):
+        if not isinstance(group, zarr.Group):
+            return False
+
+        try:
+            attrs = group.attrs["multiscales"]
+        except KeyError:
+            return False
+
+        # NOTE multiscales is designed to contain multiple multiscale-dataset, scan
+        # over them
+        name = os.path.basename(group.name)
+        for group_attrs in attrs:
+            if group_attrs["name"] == name:
+                attrs = group_attrs
+                break
+        else:
+            return False
+
+        message = None
+        try:
+            if attrs["version"] != "0.1":
+                message = "multiscale version mismatch"
+            elif attrs["name"] != group.name:
+                message = "multiscale attribute does not belong to this array"
+            elif any(info["path"] not in group for info in attrs["datasets"]):
+                message = "multiscale dataset is damaged, missing dataset"
+        except KeyError:
+            message = "multiscale dataset is damaged, missing key"
+        if message:
+            if strict:
+                raise ValueError(message)
+            else:
+                logger.warning(message)
+
+        return True
+
+    ##
+
     def _open_session(self):
         try:
             z = zarr.open(self.root_dir, mode="r")  # don't create it
@@ -408,31 +427,36 @@ class ZarrDataset(
             return False
         else:
             # verify version
+            require_version = type(self).version
             if version != type(self).version:
+                logger.debug(
+                    f"version mis-match (require: {require_version}, provide: {version})"
+                )
                 return False
 
             logger.debug(f"a uToolbox written dataset, {magic} {version}")
             return True
 
     def _enumerate_files(self):
-        #   /time/channel/setup/level
+        #   /time/channel/setup/label/level
 
-        l_str = str(self.level)
-        logger.info(f'searching level {l_str} for label "{self.label}"')
+        level_str = str(self.level)
+        logger.info(f'searching for "{self.label}" (level: {level_str})')
 
         files = []
         for t, t_root in self.handle.groups():
             for c, c_root in t_root.groups():
                 for s, s_root in c_root.groups():
-                    if l_str not in s_root:
-                        # no data for this resolution level
+                    if self.label not in s_root:
+                        # no data for this spatial setup
                         continue
-                    s_root = s_root[l_str]
-                    if self.label in s_root:
-                        # build path
-                        path = f"/{t}/{c}/{s}/{l_str}/{self.label}"
-                        files.append(path)
+                    s_root = s_root[self.label]
 
+                    path = f"/{t}/{c}/{s}/{self.label}"
+                    if level_str in s_root:
+                        # multi-level dataset
+                        path = f"{path}/{level_str}"
+                    files.append(path)
         return files
 
     def _load_array_info(self):
@@ -452,15 +476,18 @@ class ZarrDataset(
         return shape, dtype
 
     def _load_channel_info(self):
-        return self._load_injective_attributes("channel", required=True)
+        channels, self._channel_id_lut = self._load_injective_attributes(
+            "channel", required=True
+        )
+        return channels
 
     def _load_metadata(self):
         dim_info = defaultdict(lambda: defaultdict(list))
 
-        # /time[G]/channel[G]/setup[G]/level[G]/label[A]
-        indices = ["time", "channel", "setup", "level", "label"]
+        # /time[G]/channel[G]/setup[G]/label[G/A](/level[A])
+        groups = ["time", "channel", "setup", "label"]
 
-        def nested_iters(root, indices):
+        def nested_iters(root, groups):
             """
             Loop over each group and extract their raw attributes.
             
@@ -468,32 +495,54 @@ class ZarrDataset(
                 root (Group): the root to start with
                 indices (list of str): the index list to use as key in `dim_info`
             """
-            index = indices[0]
-            if len(indices) > 1:
+            if len(groups) > 1:
                 iterator = root.groups()
             else:
-                # last dimension is the array itself
-                iterator = root.arrays()
-            for name, child in iterator:
-                dim_info[index][name].append(child.attrs)  # lazy load
-                if len(indices) > 1:
-                    nested_iters(child, indices[1:])
+                iterator = chain(root.groups(), root.arrays())
 
-        nested_iters(self.handle, indices)
+            for name, child in iterator:
+                index = groups[0]
+                dim_info[index][name].append(child.attrs)
+                try:
+                    nested_iters(child, groups[1:])
+                except IndexError:
+                    pass
+
+        nested_iters(self.handle, groups)
 
         return dim_info
 
     def _load_mapped_coordinates(self):
         # NOTE we store index/coord under the same setup attrs, therefore, we construct
         # the mapped coordinate table directly
-        coords, index = [], []
-        for key, attrs in self.metadata["setup"].items():
-            for attr in attrs:
-                if "tile_coord" in attr:
-                    coords.append(attr["tile_coord"])
-                if "tile_index" in attr:
-                    index.append(attr["tile_index"])
+        coords, index, group_names = [], [], {}
+        for group_name, attrs in self.metadata["setup"].items():
+            for group_attrs in attrs:
+                if ("tile_coord" in group_attrs) and ("tile_index" in group_attrs):
+                    coords.append(group_attrs["tile_coord"])
 
+                    index_ = group_attrs["tile_index"]
+                    index.append(index_)
+                else:
+                    index_ = None
+                group_names[group_name] = index_
+
+                # ... otherwise, this setup does not belong to a tile
+
+        # save reverse lookup table
+        mapping = defaultdict(list)
+        for setup_name, index_ in group_names.items():
+            try:
+                # ensure dict is sorted in ZYX order
+                index_ = dict(sorted(index_.items(), reverse=True))
+                index_ = tuple(v for v in index_.values())
+            except AttributeError:
+                # None, probably not a tile object, therefore, no index tuple
+                index_ = None
+            mapping[index_].append(setup_name)
+        self._tile_id_lut = mapping
+
+        # build dataframe
         coords, index = pd.DataFrame(coords), pd.DataFrame(index)
 
         # rename index (which is a bit annoy to rename as multi-index)
@@ -507,6 +556,9 @@ class ZarrDataset(
 
         # build multi-index
         df = pd.concat([index, coords], axis="columns")
+        if df.empty:
+            # nothing to build `tile_coords`, use None
+            return None
         df.set_index(index.columns.to_list(), inplace=True)
 
         # rename coords
@@ -516,32 +568,44 @@ class ZarrDataset(
         return df
 
     def _load_timestamps(self) -> List[np.datetime64]:
-        return self._load_injective_attributes("time")
+        timestamps, self._timestamp_id_lut = self._load_injective_attributes(
+            "time", "timestamp"
+        )
+        return timestamps
 
     def _load_view_info(self):
-        return self._load_injective_attributes("view")
+        views, self._view_id_lut = self._load_injective_attributes("setup", "view")
+        return views
 
-    def _load_injective_attributes(self, dim_name, required=False):
+    def _load_injective_attributes(self, dim_name, key=None, required=False):
         """
-        Load group dimensional attributes that are 1-1 mapping across different group 
+        Load group dimensional attributes that are 1-1 mapping across different group
         names.
 
         Args:
             dim_name (str): dimension name to extract
+            key (str, optional): key name to search in the attributes
             required (bool, optional): the attribute must exist
+
+        Returns:
+            (Tuple[List[values], Mapping[reverse lookup]])
         """
+        key = dim_name if key is None else key
+
         mapping = defaultdict(set)
-        for key, attrs in self.metadata[dim_name].items():
+        for group_name, attrs in self.metadata[dim_name].items():
             for attr in attrs:
                 try:
-                    value = attr[dim_name]
+                    value = attr[key]
                 except KeyError:
                     if required:
-                        raise KeyError(f'"{key}" does not have attribute "{dim_name}"')
-                else:
-                    # NOTE split to try-except-else to ensure we do not create
-                    # unncessary keys
-                    mapping[key].add(value)
+                        raise KeyError(
+                            f'"{group_name}" does not have attribute "{key}"'
+                        )
+                    value = None
+                # NOTE split to try-except-else to ensure we do not create unnecessary
+                # keys
+                mapping[group_name].add(value)
 
         for key, value in mapping.items():
             inconsist = len(value) > 1
@@ -552,10 +616,22 @@ class ZarrDataset(
                 )
             mapping[key] = value
 
-        # NOTE duing `_update_inventory_index`, we rely on None to determine columns to
-        # drop
-        attrs = list(mapping.values())
-        return attrs if attrs else None
+        # NOTE during `_update_inventory_index`, we rely on `None` to determine columns
+        # to drop
+        attrs = list(set(mapping.values()))
+        if attrs and (attrs[0] is not None):
+            attrs.sort()
+        else:
+            attrs = None  # compress all the None-s into None
+
+        # generate reverse mapping (used in retrieve file list)
+        rmapping = defaultdict(list)
+        for k, v in mapping.items():
+            rmapping[v].append(k)
+        # simplify 1 item lists
+        rmapping = {k: v if len(v) > 1 else v[0] for k, v in rmapping.items()}
+
+        return attrs, rmapping
 
     def _load_voxel_size(self):
         voxel_size = set()
@@ -569,8 +645,34 @@ class ZarrDataset(
         if inconsist:
             logger.error(f"voxel size varies, using {voxel_size}")
 
+        # TODO factor in resolution level
+        logger.warning("voxel size is native resolution")
+
         return voxel_size
 
     def _retrieve_file_list(self, coord_dict):
-        print(coord_dict)  # TODO lookup attributes -> group number idF
-        raise RuntimeError("DEBUG, _retrieve_file_list")
+        # t
+        time = self._timestamp_id_lut[coord_dict.get("time", None)]
+        # c
+        channel = self._channel_id_lut[coord_dict.get("channel")]
+        # NOTE setups are mostly _not_ 1:1, therefore, we need to use `set` to filter
+        # them out
+        # s, view
+        view = self._view_id_lut[coord_dict.get("view", None)]
+        view = set(
+            [view] if isinstance(view, str) else view
+        )  # ensure it is an iterable
+        # s, tile
+        tile_index = (
+            tuple(coord_dict[index] for index in self.tile_index_str)
+            if self.tile_index_str
+            else None
+        )
+        tile = set(self._tile_id_lut[tile_index])
+        # s
+        setup = view & tile
+        assert len(setup) == 1, "unable to determine an unique setup, programmer error"
+        setup = next(iter(setup))
+
+        path = "/".join(["", time, channel, setup, self.label])
+        return path
